# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_text_main_streaming.ipynb.

# %% ../nbs/00_text_main_streaming.ipynb 3
from __future__ import annotations
from sklearn.preprocessing import MultiLabelBinarizer
from datasets import DatasetDict,Dataset,IterableDataset,load_dataset,Value
from pathlib import Path
from .utils import *
from .text_main import tokenize_function,concat_metadatas
from functools import partial
from collections import defaultdict
import warnings
from datasets.utils.logging import disable_progress_bar, enable_progress_bar

# %% auto 0
__all__ = ['TextDataControllerStreaming']

# %% ../nbs/00_text_main_streaming.ipynb 6
class TextDataControllerStreaming():
    def __init__(self,
                 inp, # HuggingFainpce Dataset or DatasetDict
                 main_text:str, # Name of the main text column
                 label_names=[], # Names of the label (dependent variable) columns
                 sup_types=[], # Type of supervised learning for each label name ('classification' or 'regression')
                 class_names_predefined=[], # List of names associated with the labels (same index order)
                 filter_dict={}, # A dictionary: {feature: filtering_function_based_on_the_feature}
                 label_tfm_dict={}, # A dictionary: {label_name: transform_function_for_that_label}
                 metadatas=[], # Names of the metadata columns
                 process_metas=True, # Whether to do simple text processing on the chosen metadatas
                 content_transformations=[], # A list of text transformations
                 content_augmentations=[], # A list of text augmentations
                 seed=None, # Random seed
                 batch_size=1024, # CPU batch size
                 num_proc=1, # Number of process for multiprocessing. This will be applied on non-streamed validation set
                 cols_to_keep=None, # Columns to keep after all processings
                 verbose=True, # Whether to print processing information
                ):
            
        self.main_text = main_text
        
        self.label_names = val2iterable(label_names)
        self.sup_types = val2iterable(sup_types)
        self._check_sup_types()
        self.label_lists = class_names_predefined
        
        self.filter_dict = filter_dict
        self.label_tfm_dict = label_tfm_dict
        self.metadatas = val2iterable(metadatas)
        self.process_metas = process_metas

        self.content_tfms = val2iterable(content_transformations)
        self.aug_tfms = val2iterable(content_augmentations)
        
        self.seed = seed
        self.is_batched = batch_size>1
        self.batch_size = batch_size
        self.num_proc = num_proc
        self.cols_to_keep = cols_to_keep

        self.main_ddict=DatasetDict()
        self.verbose = verbose
        self.verboseprint = print if verbose else lambda *a, **k: None
        if not self.verbose:
            disable_progress_bar() # turn off huggingface `map` progress bar
        else:
            enable_progress_bar()
            
        if hasattr(inp,'keys'): # is datasetdict
            if 'train' not in inp.keys(): 
                raise ValueError('The given DatasetDict has no "train" split')
            else:
                self.main_ddict['train'] = inp['train']
            val_key = list(set(inp.keys()) & set(['val','validation','valid']))
            if len(val_key)>1: raise ValueError('Your DatasetDict has more than 1 validation split')
            if len(val_key)==1:
                self.main_ddict['validation'] = inp[val_key[0]]
        else: # is dataset
            self.main_ddict['train'] = inp
          
        is_streamed=isinstance(self.main_ddict['train'],IterableDataset)
        if not is_streamed: raise Exception('This Text Data Controller only handles streamed dataset')
        
        self.all_cols = get_dset_col_names(self.main_ddict['train'])
        
        if is_streamed and ('classification' in self.sup_types) and len(self.label_lists)==0:
            raise ValueError('All classification labels must be provided when streaming')
            
        self._determine_multihead_multilabel()
        self._convert_regression_to_float()
        self._processed_call=False
        
            
    @classmethod
    def from_pickle(cls,
                    fname, # Name of the pickle file
                    parent='pickle_files' # Parent folder
                   ):
        return load_pickle(fname,parent=parent)
    
    def set_verbose(self,verbose):
        self.verbose = verbose
        self.verboseprint = print if verbose else lambda *a, **k: None
    
    def _convert_regression_to_float(self):
        if len(self.sup_types)==0: return
        # convert regression labels to float64
        reg_idxs = [i for i,v in enumerate(self.sup_types) if v=='regression']
        for i in reg_idxs:
            self.main_ddict['train'] = self.main_ddict['train'].cast_column(self.label_names[i],Value("float64"))
            if 'validation' in self.main_ddict.keys():
                self.main_ddict['validation'] = self.main_ddict['validation'].cast_column(self.label_names[i],Value("float64"))
        
    def _check_sup_types(self):
        assert len(self.label_names)==len(self.sup_types), "The number of supervised learning declaration must equal to the number of label"
        assert len(set(self.sup_types) - set(['classification','regression']))==0, 'Accepted inputs for `sup_types` are `classification` and `regression`'
        
    def _determine_multihead_multilabel(self):
        self.is_multilabel=False
        self.is_multihead=False
        if len(self.label_names)==0: return
        
        if len(self.label_names)>1:
            self.is_multihead=True
        # get label of first row
        first_label = next(iter(self.main_ddict['train']))[self.label_names[0]]
        if isinstance(first_label,(list,set,tuple)):
            # This is multi-label. Ignore self.label_names[1:]
            self.label_names = [self.label_names[0]]
            self.is_multihead=False
            self.is_multilabel=True
                     
    
    def save_as_pickles(self,
                        fname, # Name of the pickle file
                        parent='pickle_files', # Parent folder
                        drop_attributes=False # Whether to drop large-size attributes
                       ):
        if drop_attributes:
            if hasattr(self, 'main_ddict'):
                del self.main_ddict
            if hasattr(self, 'ddict_rest'):
                del self.ddict_rest
            if hasattr(self, 'aug_tfms'):
                del self.aug_tfms
        save_to_pickle(self,fname,parent=parent)
    
    def _process_metadatas(self,dtrain):
        if len(self.metadatas):
            map_func = partial(concat_metadatas,
                               main_text=self.main_text,
                               metadatas=self.metadatas,
                               process_metas=self.process_metas,
                               is_batched=self.is_batched)
            dtrain = hf_map_dset(dtrain,map_func,self.is_batched,self.batch_size,self.num_proc)
        return dtrain
    
    def _do_label_transformation(self):
        if len(self.label_names)==0 or len(self.label_tfm_dict)==0: return
        print_msg('Label Transformation',20,verbose=self.verbose)
        for f,tfm in self.label_tfm_dict.items():
            if f in self.label_names:
                _func = partial(lambda_map_batch,
                                feature=f,
                                func=tfm,
                                is_batched=self.is_batched
                               )                
                self.main_ddict['train'] = hf_map_dset(self.main_ddict['train'],_func,self.is_batched,self.batch_size,self.num_proc)
                if 'validation' in self.main_ddict.keys():
                    self.main_ddict['validation'] = hf_map_dset(self.main_ddict['validation'],
                                                                _func,
                                                                self.is_batched,
                                                                self.batch_size,
                                                                self.num_proc)
        self.verboseprint('Done')
                    
                      
                
    def _create_label_mapping_func(self,encoder_classes):
        if self.is_multihead:
            label2idxs = [{v:i for i,v in enumerate(l_classes)} for l_classes in encoder_classes]
            _func = lambda inp: {'label': [[label2idxs[i][v] if len(label2idxs[i]) else v for i,v in enumerate(vs)] \
                                           for vs in zip(*[inp[l] for l in self.label_names])] if self.is_batched \
                                 else [label2idxs[i][v] if len(label2idxs[i]) else v for i,v in enumerate([inp[l] for l in self.label_names])]
                                }
            
        else: # single-head
            if self.sup_types[0]=='regression':
                _func1 = lambda x: x
            else:
                label2idx = {v:i for i,v in enumerate(encoder_classes[0])}
                _func1 = lambda x: label2idx[x]
                
            _func = partial(lambda_map_batch,
                           feature=self.label_names[0],
                           func=_func1,
                           output_feature='label',
                           is_batched=self.is_batched)
        return _func
        
    def _encode_labels(self):
        if len(self.label_names)==0: return
        print_msg('Label Encoding',verbose=self.verbose)
        
        if len(self.label_lists) and not isinstance(self.label_lists[0],list):
            self.label_lists = [self.label_lists]
                    
        encoder_classes=[]
        if not self.is_multilabel:
            for idx,l in enumerate(self.label_names):
                if self.sup_types[idx]=='regression':
                    l_classes=[]
                else: # classification
                    l_classes = sorted(list(self.label_lists[idx]))
                encoder_classes.append(l_classes)
                
            _func = self._create_label_mapping_func(encoder_classes)
            
            self.main_ddict['train'] = hf_map_dset(self.main_ddict['train'],_func,self.is_batched,self.batch_size,self.num_proc)
            if 'validation' in self.main_ddict.keys():
                self.main_ddict['validation'] = hf_map_dset(self.main_ddict['validation'],_func,self.is_batched,self.batch_size,self.num_proc)
                    
        else:
            # For MultiLabel, we transform the label itself to one-hot (or actually, few-hot)
            l_classes = sorted(list(self.label_lists[0]))   
            encoder_classes.append(l_classes)
            
            l_encoder = MultiLabelBinarizer(classes=encoder_classes[0])
            _ = l_encoder.fit(None)
            _func = partial(lambda_map_batch,
                            feature=self.label_names[0],
                            func=lambda x: l_encoder.transform(x),
                            output_feature='label',
                            is_batched=self.is_batched,
                            is_func_batched=True)
            self.main_ddict['train'] = hf_map_dset(self.main_ddict['train'],_func,self.is_batched,self.batch_size,self.num_proc)
            if 'validation' in self.main_ddict.keys():
                self.main_ddict['validation'] = hf_map_dset(self.main_ddict['validation'],_func,self.is_batched,self.batch_size,self.num_proc)
            
        self.label_lists = encoder_classes
        self.verboseprint('Done')
        
            
            
    def _simplify_ddict(self):
        print_msg('Dropping unused features',20,verbose=self.verbose)
        if self.cols_to_keep is None:
            self.cols_to_keep= [self.main_text] + self.metadatas + self.label_names
        cols_to_remove = set(self.all_cols) - set(self.cols_to_keep)
        self.main_ddict['train']=self.main_ddict['train'].remove_columns(list(cols_to_remove))
        if 'validation' in self.main_ddict.keys():
            self.main_ddict['validation']=self.main_ddict['validation'].remove_columns(list(cols_to_remove))
        self.verboseprint('Done')

    def _do_filtering(self,dtrain):
        if len(self.filter_dict):
            col_names = get_dset_col_names(dtrain)
            for f,tfm in self.filter_dict.items():
                if f in col_names:
                    _func = partial(lambda_batch,
                                    feature=f,
                                    func=tfm,
                                    is_batched=self.is_batched)
                    dtrain = hf_filter_dset(dtrain,_func,self.is_batched,self.batch_size,self.num_proc)
        return dtrain
        

    def _do_transformation_tokenization(self,dtrain):
        tok_func = partial(tokenize_function,tok=self.tokenizer,max_length=self.max_length)
        if len(self.content_tfms):            
            for tfm in self.content_tfms:
                _func = partial(lambda_map_batch,
                                feature=self.main_text,
                                func=tfm,
                                is_batched=self.is_batched)
                dtrain = hf_map_dset(dtrain,_func,self.is_batched,self.batch_size,self.num_proc)
        
        _func = partial(lambda_map_batch,
                        feature=self.main_text,
                        func=tok_func,
                        output_feature=None,
                        is_batched=self.is_batched)
        dtrain = hf_map_dset(dtrain,_func,self.is_batched,self.batch_size,self.tok_num_proc)
            
        return dtrain 

    def _do_transformation_augmentation_tokenization(self,dtrain,tok_func,all_tfms):

        num_proc=1 # high num_proc is not beneficial with each batch size (which is only around 1k)
        # Content transformation + augmentation
        for tfm in all_tfms:
            bs = self.batch_size
            is_func_batched=False
            is_batched = self.is_batched
            if hasattr(tfm, "run_on_gpu") and getattr(tfm,'run_on_gpu')==True:
                bs = min(32,self.batch_size) if not hasattr(tfm, "batch_size") else getattr(tfm,'batch_size')
                is_func_batched=True
                is_batched=True

            _func = partial(lambda_map_batch,
                            feature=self.main_text,
                            func=tfm,
                            is_batched=is_batched,
                            is_func_batched=is_func_batched
                            )
            dtrain = hf_map_dset(dtrain,_func,
                                 is_batched=is_batched,
                                 batch_size=bs,
                                 num_proc=num_proc
                                )
        # Tokenization
        _func = partial(lambda_map_batch,
                        feature=self.main_text,
                        func=tok_func,
                        output_feature=None,
                        is_batched=self.is_batched)
        dtrain = hf_map_dset(dtrain,_func,self.is_batched,self.batch_size,num_proc)
            
        return dtrain
    
    def _construct_generator_with_batch(self,dset,tok_func,all_tfms):        
        def _get_generator(dset):
            for v in dset: yield v
            
        final_dict = defaultdict(list)
        for inp in dset: # dset is generator
            # inp[text_name] will be a single item
            for k,v in inp.items():
                final_dict[k].append(v)
            
            if len(final_dict[self.main_text])==self.batch_size:
                # a full batch (self.batch_size) is created
                dtrain = Dataset.from_dict(final_dict)
                dtrain = self._do_transformation_augmentation_tokenization(dtrain,tok_func,all_tfms)
                yield from _get_generator(dtrain)
                final_dict=defaultdict(list)            
            
        if len(final_dict[self.main_text]):
            # hasn't reached batch_size (of last batch)
            dtrain = Dataset.from_dict(final_dict)
            dtrain = self._do_transformation_augmentation_tokenization(dtrain,tok_func,all_tfms)
            yield from _get_generator(dtrain)
            
        
            
    def _do_transformation_augmentation_tokenization_generator(self):
        tok_func = partial(tokenize_function,tok=self.tokenizer,max_length=self.max_length)
        all_tfms = self.content_tfms + self.aug_tfms
        
        self.main_ddict['train'] = IterableDataset.from_generator(self._construct_generator_with_batch,
                                                   gen_kwargs={'dset': self.main_ddict['train'],
                                                               'tok_func':tok_func,
                                                               'all_tfms': all_tfms
                                                              }
                                                                 )
    
    def _do_transformation_augmentation_tokenization_generator_linebyline(self):
        def _get_generator(dset,tok_func,all_tfms):
            for inp in dset:
                # inp[text_name] will be a single item
                inp[self.main_text]=all_tfms(inp[self.main_text])
                result_dict = tok_func(inp[self.main_text])
                for k,v in result_dict.items():
                    inp[k]=v
                yield inp
        
        # no padding for tokenization
        tok_func = partial(tokenize_function,tok=self.tokenizer,max_length=-1) 
        all_tfms = self.content_tfms + self.aug_tfms
        all_tfms = partial(func_all,functions=all_tfms) if len(all_tfms) else lambda x: x
           
        self.main_ddict['train'] = IterableDataset.from_generator(_get_generator,
                                                   gen_kwargs={'dset': self.main_ddict['train'],
                                                               'tok_func':tok_func,
                                                               'all_tfms': all_tfms
                                                              }
                                                                 )

        
    def process_and_tokenize(self,
                             tokenizer, # Tokenizer (preferably from HuggingFace)
                             max_length=None, # pad to model's allowed max length (default is max_sequence_length)
                             tok_num_proc=None, # Number of processes for tokenization
                             line_by_line=False, # To whether process + tokenize each sentence separately. Faster, but no padding applied
                            ):
        if self._processed_call:
            warnings.warn('Your dataset has already been processed. Returning the previous processed DatasetDict...')
            return self.main_ddict
        
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.tok_num_proc = tok_num_proc if tok_num_proc else self.num_proc
        
        # Filtering
        print_msg('Data Filtering',20,verbose=self.verbose)
        for k in self.main_ddict.keys():   
            self.main_ddict[k] = self._do_filtering(self.main_ddict[k])
        self.verboseprint('Done')

        
        # Process metadatas
        print_msg('Metadata Simple Processing & Concatenating to Main Content',verbose=self.verbose)
        for k in self.main_ddict.keys():   
            self.main_ddict[k] = self._process_metadatas(self.main_ddict[k])
        self.verboseprint('Done')
        
        # Label transformation
        self._do_label_transformation()
        
        # Process labels
        self._encode_labels()

        # Dropping unused columns
        self._simplify_ddict()

        if self.seed:
            seed_everything(self.seed)
            
        # Content transformation + tokenization for validation
        if 'validation' in self.main_ddict.keys():
            print_msg('Performing Content Transformation and Tokenization on validation set',verbose=self.verbose)
            self.main_ddict['validation'] = self._do_transformation_tokenization(self.main_ddict['validation'])
            self.verboseprint('Done')
 
        # Content transformation + augmentation + tokenization for train
        print_msg('Creating a generator for content transformation, augmentation and tokenization on train set',verbose=self.verbose)
        if line_by_line:
            self._do_transformation_augmentation_tokenization_generator_linebyline()
        else:
            self._do_transformation_augmentation_tokenization_generator()
        self.verboseprint('Done')
        
        self._processed_call=True
    
        
    
    def set_data_collator(self,data_collator):
        self.data_collator = data_collator
        
    
    def prepare_test_dataset_from_csv(self,
                                      file_path, # path to csv file
                                      do_filtering=False # whether to perform data filtering on this test set
                                     ):
        file_path = Path(file_path)
        ds = load_dataset(str(file_path.parent),
                          data_files=file_path.name,
                          split='train')
        return self.prepare_test_dataset(ds,do_filtering)
    
    def prepare_test_dataset_from_df(self,
                                     df, # Pandas Dataframe
                                     validate=True, # whether to perform input data validation
                                     do_filtering=False # whether to perform data filtering on this test set 
                                    ):
        if validate:
            check_input_validation(df)
        ds = Dataset.from_pandas(df)
        return self.prepare_test_dataset(ds,do_filtering)
    
    def prepare_test_dataset_from_raws(self,
                                       content, # Either a single sentence, list of sentence or a dictionary with keys are metadata columns and values are list
                                      ):
        if len(self.metadatas) and not isinstance(content,dict):
            raise ValueError(f'There is/are metadatas in the preprocessing step. Please include a dictionary including these keys for metadatas: {self.metadatas}, and texture content: {self.main_text}')
            
        _dic = {self.main_text:[content]} if isinstance(content,str) else content
        for k in _dic.keys():
            _dic[k] = val2iterable(_dic[k])
        
        test_dict = Dataset.from_dict(_dic)
        
        # set num_proc to 1 for small data processing
        _tmp1 = self.num_proc
        _tmp2 = self.tok_num_proc
        self.num_proc=1
        self_tok_num_proc=1
        results = self.prepare_test_dataset(test_dict,do_filtering=False)
        self.num_proc = _tmp1
        self.tok_num_proc=_tmp2
        
        return results
    
    def prepare_test_dataset(self,
                             test_dset, # The HuggingFace Dataset as Test set
                             do_filtering=False, # whether to perform data filtering on this test set
                            ):
        test_cols = set(get_dset_col_names(test_dset))
        label_names_set = set(self.label_names)
        test_cols = test_cols - label_names_set
        missing_cols = set(self.cols_to_keep) - label_names_set - test_cols
        if len(missing_cols):
            raise ValueError(f'Test set does not have these columns required for preprocessings: {missing_cols}')
            
        print_msg('Start Test Set Transformation',20,verbose=self.verbose)

        # Filtering
        if do_filtering:
            print_msg('Data Filtering',20,verbose=self.verbose)
            test_dset = self._do_filtering(test_dset)
            self.verboseprint('Done')
        
        # Process metadatas
        print_msg('Metadata Simple Processing & Concatenating to Main Content',verbose=self.verbose)    
        test_dset = self._process_metadatas(test_dset)
        self.verboseprint('Done')
        
        # Drop unused columns
        print_msg('Dropping unused features',20,verbose=self.verbose)
        cols_to_remove = test_cols - set(self.cols_to_keep)
        test_dset=test_dset.remove_columns(list(cols_to_remove))
        self.verboseprint('Done')
        
        
        # Content transformation and tokenization
        print_msg('Performing Content Transformation and Tokenization on test set',verbose=self.verbose)
        test_dset = self._do_transformation_tokenization(test_dset)
        self.verboseprint('Done')
        
        return test_dset




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/05_models.roberta.deep_hierarchical_classifiers.ipynb.

# %% ../../../nbs/05_models.roberta.deep_hierarchical_classifiers.ipynb 3
from __future__ import annotations
import torch
from transformers.models.roberta.configuration_roberta import RobertaConfig
from transformers.models.roberta.modeling_roberta import RobertaModel
from transformers.modeling_outputs import SequenceClassifierOutput
from transformers.models.roberta.modeling_roberta import RobertaPreTrainedModel
from sklearn.preprocessing import MultiLabelBinarizer

# %% auto 0
__all__ = ['build_DHC_conditional_mask', 'loss_for_DHC', 'RobertaConcatHeadDHCRoot', 'RobertaSimpleHSCDHCSequenceClassification',
           'RobertaHSCDHCSequenceClassification']

# %% ../../../nbs/05_models.roberta.deep_hierarchical_classifiers.ipynb 5
def build_DHC_conditional_mask(df_labels,
                                  label1,label2):
    """
    This is really similar to `build_standard_condition_mask`, 
    but we don't concatenate l1 1-hot matrix and l2 1-hot matrix
    """
    df_labels = df_labels.drop_duplicates().sort_values([label1,label2])
    _d = df_labels.groupby([label1])[label2].apply(list).to_dict()
    
    mlb = MultiLabelBinarizer()
    results = mlb.fit_transform([v for k,v in sorted(_d.items())])
    return torch.from_numpy(results).float()

# %% ../../../nbs/05_models.roberta.deep_hierarchical_classifiers.ipynb 10
def _check_hierarchy(l1_pred,l2_pred,dhc_mask,size_l2):
    # Check if the predicted class at L2 is a children of the class predicted at L1 for the entire batch
    # if is a children return 0, else 1
    _l1_pred = dhc_mask[l1_pred] # bs,size_l2
    _l2_pred = torch.nn.functional.one_hot(l2_pred, num_classes=size_l2) # bs, size_l2
    return (~torch.mul(_l1_pred,_l2_pred).sum(axis=1).bool()).float()

# %% ../../../nbs/05_models.roberta.deep_hierarchical_classifiers.ipynb 11
def loss_for_DHC(l1_repr_logits, # Head 1's logit output
                 l2_repr_logits, # Head 2's logit output
                 labels_l1, # True label for head 1
                 labels_l2, # True label for head 2
                 dhc_mask, # A one-hot matrix between classes of head 1 and 2
                 lloss_weight=1.0, # Weight for Layer Loss (lloss)
                 dloss_weight=0.8 # Weight for Dependence Loss (dloss)
                ):
    "Reference: https://github.com/Ugenteraan/Deep_Hierarchical_Classification/blob/main/model/hierarchical_loss.py"
    size_l2 = dhc_mask.shape[1]
    
    # calculate lloss
    lloss = 0
    lloss += torch.nn.CrossEntropyLoss()(l1_repr_logits, labels_l1)
    lloss += torch.nn.CrossEntropyLoss()(l2_repr_logits, labels_l2)
    lloss = lloss_weight*lloss


    # calculate dloss
    l1_pred = torch.argmax(l1_repr_logits, dim=1)
    l2_pred = torch.argmax(l2_repr_logits, dim=1)
    D_l = _check_hierarchy(l1_pred,l2_pred,dhc_mask,size_l2)
    l1_not_match_true_label = (~(l1_pred == labels_l1)).float()
    l2_not_match_true_label = (~(l2_pred == labels_l2)).float()
    dloss_l1 = torch.nn.CrossEntropyLoss()(l1_repr_logits*D_l[:,None]*l1_not_match_true_label[:,None], labels_l1)
    dloss_l2 = torch.nn.CrossEntropyLoss()(l2_repr_logits*D_l[:,None]*l2_not_match_true_label[:,None], labels_l2)
    dloss = dloss_weight*(dloss_l1 + dloss_l2)  
    
    
    loss = lloss + dloss
    
    return loss

# %% ../../../nbs/05_models.roberta.deep_hierarchical_classifiers.ipynb 13
class RobertaConcatHeadDHCRoot(torch.nn.Module):
    """
    Concatenated head for Roberta DHC Classification Model. 
    """
    def __init__(self,
                 config, # HuggingFace model configuration
                 classifier_dropout=0.1, # Dropout ratio (for dropout layer right before the last nn.Linear)
                 last_hidden_size=768, # Last hidden size (before the last nn.Linear)
                 layer2concat=4, # number of hidden layer to concatenate (counting from top)
                 **kwargs
                ):
        super().__init__()
        self.classifier_dropout=classifier_dropout
        self.last_hidden_size=last_hidden_size
        self.dropout = torch.nn.Dropout(classifier_dropout)
        self.pre_classifier = torch.nn.Linear(layer2concat*config.hidden_size,
                                              last_hidden_size)    
    def forward(self, inp, **kwargs):
        x = inp
        x = self.dropout(x)
        x = self.pre_classifier(x)
        x = torch.tanh(x)
#         x = torch.relu(x)
        return x

# %% ../../../nbs/05_models.roberta.deep_hierarchical_classifiers.ipynb 15
class RobertaSimpleHSCDHCSequenceClassification(RobertaPreTrainedModel):
    """
    Roberta Simple-DHC Architecture with Hidden-State-Concatenation for Sequence Classification task
    """
    config_class = RobertaConfig

    def __init__(self, 
                 config, # HuggingFace model configuration
                 dhc_mask, # A one-hot matrix between classes of head 1 and 2
                 lloss_weight=1.0, # Weight for Layer Loss (lloss)
                 dloss_weight=0.8, # Weight for Dependence Loss (dloss)
                 layer2concat=4, # number of hidden layer to concatenate (counting from top)
                 device=None # CPU or GPU
                ):
        
        super().__init__(config)
        self.training_device = device if device is not None else torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.size_l1 = dhc_mask.shape[0]
        self.size_l2 = dhc_mask.shape[1]
        self.head_class_sizes=[self.size_l1,self.size_l2] # will be useful for metric calculation later

        self.dhc_mask = dhc_mask.to(self.training_device)
        self.layer2concat=layer2concat
    
        self.lloss_weight = lloss_weight
        self.dloss_weight = dloss_weight
        
        self.body_model = RobertaModel(config, add_pooling_layer=False)
        
        self.linear_L1_logit = torch.nn.Linear(layer2concat*config.hidden_size,self.size_l1)

        self.linear_L2 = torch.nn.Linear(layer2concat*config.hidden_size, abs(self.size_l2-self.size_l1))
        self.linear_L2_logit = torch.nn.Linear(self.size_l2,self.size_l2)
        
    
    def forward(self, input_ids=None, attention_mask=None, token_type_ids=None,
                labels=None, 
                output_attentions= None,
                output_hidden_states= None,
                return_dict= None,
                **kwargs):
        # Use model body to get encoder representations
        # the only ones we need for now are input_ids and attention_mask
        output_hidden_states = output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        outputs = self.body_model(input_ids, 
                                  attention_mask=attention_mask,
                                  token_type_ids=token_type_ids, 
                                  output_attentions=output_attentions,
                                  output_hidden_states=output_hidden_states,
                                  return_dict=return_dict,
                                  **kwargs)
        
        hidden_states = outputs['hidden_states'] # tuples with len 13 (number of layer/block)
        # each with shape: (bs,seq_len,hidden_size_len), e.g. for phobert: (bs,256, 768)
        # Note: hidden_size_len = embedding_size
        
        hidden_concat = torch.cat([hidden_states[i][:,0] for i in range(-1,-self.layer2concat-1,-1)],
                                  -1) # (bs,768*4)
        
        # root_representation
        root_repr = hidden_concat # (bs,768*4) 
        
        l1_repr_logits = self.linear_L1_logit(root_repr) # (bs,size_l1)
        
        l2_repr = self.linear_L2(root_repr) # (bs,size_l2-size_l1)
        l2_repr = torch.cat((l2_repr,l1_repr_logits),dim=1) # (bs, size_l2)
        l2_repr_logits = self.linear_L2_logit(l2_repr) # (bs, size_l2) 
        
        loss = None
        
        if labels is not None:
            # labels shape: (bs,2), first is L1, second is L2
            labels_l1 = labels[:,0].view(-1) #(bs,)
            labels_l2 = labels[:,1].view(-1) #(bs,)

            loss = loss_for_DHC(l1_repr_logits, l2_repr_logits, 
                                labels_l1, labels_l2, 
                                self.dhc_mask,
                                self.lloss_weight,self.dloss_weight)
            
        # Return model output object
        return SequenceClassifierOutput(loss=loss, logits=(l1_repr_logits,l2_repr_logits),
                                     hidden_states=None,
                                     attentions=outputs.attentions)

# %% ../../../nbs/05_models.roberta.deep_hierarchical_classifiers.ipynb 17
class RobertaHSCDHCSequenceClassification(RobertaPreTrainedModel):
    """
    Roberta DHC Architecture with Hidden-State-Concatenation for Sequence Classification task
    """    
    config_class = RobertaConfig

    def __init__(self, 
                 config, # HuggingFace model configuration
                 dhc_mask, # A one-hot matrix between classes of head 1 and 2
                 classifier_dropout=0.1, # Dropout ratio (for dropout layer right before the last nn.Linear)
                 last_hidden_size=768, # Last hidden size (before the last nn.Linear)
                 linear_l1_size=None, # last hidden size for head 1
                 linear_l2_size=None, # last hidden size for head 2
                 lloss_weight=1.0, # Weight for Layer Loss (lloss)
                 dloss_weight=0.8, # Weight for Dependence Loss (dloss)
                 layer2concat=4, # number of hidden layer to concatenate (counting from top)
                 device=None # CPU or GPU
                ):
        
        super().__init__(config)
        self.training_device = device if device is not None else torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.size_l1 = dhc_mask.shape[0]
        self.size_l2 = dhc_mask.shape[1]
        self.head_class_sizes=[self.size_l1,self.size_l2] # will be useful for metric calculation later

        self.dhc_mask = dhc_mask.to(self.training_device)
        self.layer2concat=layer2concat

        
        if linear_l1_size is None: linear_l1_size = (last_hidden_size+self.size_l1)//2 # 389
        if linear_l2_size is None: linear_l2_size = (last_hidden_size+self.size_l2)//2 # 417
            
        self.lloss_weight = lloss_weight
        self.dloss_weight = dloss_weight
        
        self.body_model = RobertaModel(config, add_pooling_layer=False)
        self.root_representation = RobertaConcatHeadDHCRoot(config=config,
                                                            classifier_dropout=classifier_dropout,
                                                            last_hidden_size=last_hidden_size)
        
        self.linear_L1 = torch.nn.Linear(last_hidden_size, linear_l1_size)
        self.linear_L2 = torch.nn.Linear(last_hidden_size, linear_l2_size)
        
        self.linear_L1_logit = torch.nn.Linear(linear_l1_size,self.size_l1)
        self.linear_L2_logit = torch.nn.Linear(linear_l1_size + linear_l2_size,self.size_l2)
        

    
    def forward(self, input_ids=None, attention_mask=None, token_type_ids=None,
                labels=None, 
                output_attentions= None,
                output_hidden_states= None,
                return_dict= None,
                **kwargs):
        # Use model body to get encoder representations
        # the only ones we need for now are input_ids and attention_mask
        output_hidden_states = output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        outputs = self.body_model(input_ids, 
                                  attention_mask=attention_mask,
                                  token_type_ids=token_type_ids, 
                                  output_attentions=output_attentions,
                                  output_hidden_states=output_hidden_states,
                                  return_dict=return_dict,
                                  **kwargs)
        
        hidden_states = outputs['hidden_states'] # tuples with len 13 (number of layer/block)
        # each with shape: (bs,seq_len,hidden_size_len), e.g. for phobert: (bs,256, 768)
        # Note: hidden_size_len = embedding_size
        
        hidden_concat = torch.cat([hidden_states[i][:,0] for i in range(-1,-self.layer2concat-1,-1)],
                                  -1) # (bs,768*4)
        
        # root_representation
        root_repr = self.root_representation(hidden_concat) # (bs,768) 
        
        l1_repr = self.linear_L1(root_repr) # (bs,linear_L1_size)
        l1_repr_logits = self.linear_L1_logit(l1_repr) # (bs,size_l1)
        l2_repr = self.linear_L2(root_repr) # (bs,linear_L2_size)
        l2_repr = torch.cat((l2_repr,l1_repr),dim=1) # (bs, linear_l1_size + linear_l2_size)
        l2_repr_logits = self.linear_L2_logit(l2_repr) # (bs, size_l2) 

        
        loss = None
        
        if labels is not None:
            # labels shape: (bs,2), first is L3, second is L4
            labels_l1 = labels[:,0].view(-1) #(bs,)
            labels_l2 = labels[:,1].view(-1) #(bs,)

            loss = loss_for_DHC(l1_repr_logits, l2_repr_logits, 
                                labels_l1, labels_l2, 
                                self.dhc_mask,
                                self.lloss_weight,self.dloss_weight)
            
        # Return model output object
        return SequenceClassifierOutput(loss=loss, logits=(l1_repr_logits,l2_repr_logits),
                                     hidden_states=None,
                                     attentions=outputs.attentions)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/09_models.gpt2.classifiers.ipynb.

# %% ../../../nbs/09_models.gpt2.classifiers.ipynb 3
from __future__ import annotations
import torch
from transformers.models.gpt2.modeling_gpt2 import GPT2Model,GPT2PreTrainedModel
from transformers.models.gpt2.configuration_gpt2 import GPT2Config
from transformers.modeling_outputs import SequenceClassifierOutputWithPast
from ...model_main import loss_for_classification
from ...utils import *

# %% auto 0
__all__ = ['GPT2BaseForSequenceClassification', 'GPT2HiddenStateConcatForSequenceClassification']

# %% ../../../nbs/09_models.gpt2.classifiers.ipynb 5
class GPT2BaseForSequenceClassification(GPT2PreTrainedModel):
    """
    GPT2 Architecture for Sequence Classification task
    Based on: https://github.com/huggingface/transformers/blob/main/src/transformers/models/gpt2/modeling_gpt2.py#L1376
    """
    config_class = GPT2Config

    def __init__(self,
                 config, # HuggingFace model configuration
                 is_multilabel=False, # Whether this is a multilabel classification
                 is_multihead=False, # Whether this is a multihead (multi-level) classification
                 head_class_sizes=[], # Class size for each head
                 head_weights=[], # loss weight for each head. This will be multiplied to the loss of each head's output
                 head_class=None, # The class object of the head. 
                 **head_class_kwargs, # Keyword arguments for the head class
                ):
        super().__init__(config)
        self.is_multilabel = is_multilabel
        self.is_multihead = is_multihead
        self.head_class_sizes = val2iterable(head_class_sizes)
        self.head_weights = val2iterable(head_weights,lsize=len(self.head_class_sizes))
        
        # set num_labels for config
        num_labels = sum(self.head_class_sizes)
        config.num_labels = num_labels
        
        self.body_model = GPT2Model(config)
        
        # Set up token classification head
        if head_class is None:
            self.head = torch.nn.Linear(config.n_embd, num_labels, bias=False)
        else:
            self.head = head_class(config,**head_class_kwargs)

        
        # Model parallel
        self.model_parallel = False
        self.device_map = None
        
        self.config.pad_token_id = self.config.eos_token_id
        
    def forward(
        self,
        input_ids= None,
        past_key_values= None,
        attention_mask= None,
        token_type_ids= None,
        position_ids= None,
        head_mask= None,
        inputs_embeds= None,
        labels= None,
        use_cache= None,
        output_attentions= None,
        output_hidden_states= None,
        return_dict= None,
    ):
        # Use model body to get encoder representations
        # the only ones we need for now are input_ids and attention_mask
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict
        output_hidden_states = output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states

        outputs = self.body_model(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
              
        sequence_output = outputs[0] # last hidden state: (bs,sequence_length,hidden_size: 768)
        
        # get the idx of the last token (typically just -1), to be used for classification
        if input_ids is not None:
            batch_size, sequence_length = input_ids.shape[:2]
        else:
            batch_size, sequence_length = inputs_embeds.shape[:2]

        assert (
            self.config.pad_token_id is not None or batch_size == 1
        ), "Cannot handle batch sizes > 1 if no padding token is defined."
        if self.config.pad_token_id is None:
            sequence_lengths = -1
        else:
            if input_ids is not None:
                sequence_lengths = (torch.eq(input_ids, self.config.pad_token_id).long().argmax(-1) - 1).to(
                    sequence_output.device
                )
            else:
                sequence_lengths = -1
                print(
                    f"{self.__class__.__name__} will not detect padding tokens in `inputs_embeds`. Results may be "
                    "unexpected if using padding tokens in conjunction with `inputs_embeds.`"
                )
                
        # sequence length at this point is just the idx (or indices) of the last token        
        sequence_output = sequence_output[torch.arange(batch_size, device=sequence_output.device), 
                                          sequence_lengths,:] # (bs,hidden_sizes)
        
        logits = self.head(sequence_output) # (bs,sum of all class sizes)
        
        # Calculate losses
        if labels is None:
            loss=None
        else:            
            loss = loss_for_classification(logits, labels, 
                                   self.is_multilabel,
                                   self.is_multihead, 
                                   self.head_class_sizes,
                                   self.head_weights)
            
        if not return_dict:
            output = (logits,) + outputs[1:]
            return ((loss,) + output) if loss is not None else output
        
        # Return model output object
        
        return SequenceClassifierOutputWithPast(
            loss=loss,
            logits=logits,
            past_key_values=outputs.past_key_values,
            hidden_states=None,
            attentions=outputs.attentions,
        )

# %% ../../../nbs/09_models.gpt2.classifiers.ipynb 7
class GPT2HiddenStateConcatForSequenceClassification(GPT2PreTrainedModel):
    """
    GPT2 Architecture for Sequence Classification task
    Based on: https://github.com/huggingface/transformers/blob/main/src/transformers/models/gpt2/modeling_gpt2.py#L1376
    """
    config_class = GPT2Config

    def __init__(self,config, # HuggingFace model configuration
                 layer2concat=4, # number of hidden layer to concatenate (counting from top)
                 is_multilabel=False, # Whether this is a multilabel classification
                 is_multihead=False, # Whether this is a multihead (multi-level) classification
                 head_class_sizes=[], # Class size for each head
                 head_weights=[], # loss weight for each head. This will be multiplied to the loss of each head's output
                 head_class=None, # The class object of the head
                 **head_class_kwargs, # Keyword arguments for the head class
                ):
        super().__init__(config)
        self.is_multilabel = is_multilabel
        self.is_multihead = is_multihead
        self.head_class_sizes = val2iterable(head_class_sizes)
        self.head_weights = val2iterable(head_weights,lsize=len(self.head_class_sizes))
        self.layer2concat=layer2concat

        # set num_labels for config
        num_labels = sum(self.head_class_sizes)
        config.num_labels = num_labels
        
        self.body_model = GPT2Model(config)
        
        # Set up classification head
        self.head = head_class(config=config,layer2concat=layer2concat,
                                              **head_class_kwargs)

        # Model parallel
        self.model_parallel = False
        self.device_map = None

        self.config.pad_token_id = self.config.eos_token_id

    def forward(
        self,
        input_ids= None,
        past_key_values= None,
        attention_mask= None,
        token_type_ids= None,
        position_ids= None,
        head_mask= None,
        inputs_embeds= None,
        labels= None,
        use_cache= None,
        output_attentions= None,
        output_hidden_states= None,
        return_dict= None):
        
        
        # Use model body to get encoder representations
        # the only ones we need for now are input_ids and attention_mask
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict
        output_hidden_states = output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        
        outputs = self.body_model(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
              
        sequence_output = outputs[0] # (bs,sequence_length,hidden_size)
        hidden_states = outputs['hidden_states'] # tuples with 12 layers
        
        # get the idx of the last token (typically just -1), to be used for classification
        if input_ids is not None:
            batch_size, sequence_length = input_ids.shape[:2]
        else:
            batch_size, sequence_length = inputs_embeds.shape[:2]

        assert (
            self.config.pad_token_id is not None or batch_size == 1
        ), "Cannot handle batch sizes > 1 if no padding token is defined."
        if self.config.pad_token_id is None:
            sequence_lengths = -1
        else:
            if input_ids is not None:
#                 sequence_lengths = torch.ne(input_ids, self.config.pad_token_id).sum(-1) - 1
                sequence_lengths = (torch.eq(input_ids, self.config.pad_token_id).long().argmax(-1) - 1).to(
                                    sequence_output.device
                                )
            else:
                sequence_lengths = -1
                print(
                    f"{self.__class__.__name__} will not detect padding tokens in `inputs_embeds`. Results may be "
                    "unexpected if using padding tokens in conjunction with `inputs_embeds.`"
                )
        # sequence length at this point is just the idx (or indices) of the last token        
        
        hidden_concat = torch.cat([hidden_states[i][torch.arange(batch_size, device=sequence_output.device), sequence_lengths,:] for i in range(-1,-self.layer2concat-1,-1)],
                                  -1)        
        logits = self.head(hidden_concat) # (bs,sum of all class sizes)
        
    
        # Calculate losses
        if labels is None:
            loss=None
        else:
            loss = loss_for_classification(logits, labels, 
                                   self.is_multilabel,
                                   self.is_multihead, 
                                   self.head_class_sizes,
                                   self.head_weights)
            
        if not return_dict:
            output = (logits,) + outputs[1:]
            return ((loss,) + output) if loss is not None else output
        
        # Return model output object
        return SequenceClassifierOutputWithPast(
            loss=loss,
            logits=logits,
            past_key_values=outputs.past_key_values,
            hidden_states=None,
            attentions=outputs.attentions,
        )
    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_model_lm_main.ipynb.

# %% ../nbs/03_model_lm_main.ipynb 3
from __future__ import annotations
import os
from transformers import Trainer, TrainingArguments, AutoConfig
from datasets import DatasetDict,Dataset
import torch
import gc
import math
from functools import partial
import evaluate
from .utils import *
from .text_main_lm import TextDataLMController
from .text_main_lm_streaming import TextDataLMControllerStreaming
from transformers import pipeline

# %% auto 0
__all__ = ['language_model_init', 'compute_lm_accuracy', 'preprocess_lm_logits_for_metrics', 'finetune_lm',
           'extract_hidden_states', 'ModelLMController']

# %% ../nbs/03_model_lm_main.ipynb 5
def language_model_init(model_class, # Model's class object, e.g. AutoModelForMaskedLM
                        cpoint_path=None, # Either model string name on HuggingFace, or the path to model checkpoint. Put `None` to train from scratch
                        config=None, # Model config. If not provided, AutoConfig is used to load config from cpoint_path                        
                        device=None, # Device to train on
                        seed=None, # Random seed
                       ):
    """
    To initialize a language model, either masked or casual
    """
    if device is None: device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    if config is None:
        config = AutoConfig.from_pretrained(cpoint_path)

    if seed:
        seed_everything(seed)
    if cpoint_path:
        model = model_class.from_pretrained(cpoint_path,config=config).to(device)
    else:
        print('Initiate a new language model from scratch')
        model = model_class.from_config(config)
    
    print(f'Total parameters: {sum(param.numel() for param in model.parameters())}')
    print(f'Total trainable parameters: {sum(param.numel() for param in model.parameters() if param.requires_grad)}')
    
    return model

# %% ../nbs/03_model_lm_main.ipynb 11
def compute_lm_accuracy(eval_preds, # An EvalPrediction object from HuggingFace 
                        is_mlm, # if this is masked language model, set to `True`. If this is casual language model, set to `False`
                     ):
    
    """    
    Reference: https://github.com/huggingface/transformers/blob/main/examples/pytorch/language-modeling/run_mlm.py#L592C35-L592C35
    """
    metric = evaluate.load("accuracy")
    preds, labels = eval_preds
    # preds have the same shape as the labels, after the argmax(-1) has been calculated
    # by preprocess_logits_for_metrics
    if is_mlm:
        labels = labels.reshape(-1)
        preds = preds.reshape(-1)
        mask = labels != -100
        labels = labels[mask]
        preds = preds[mask]
        return metric.compute(predictions=preds, references=labels)
    else:
        # we need to shift the labels
        labels = labels[:, 1:].reshape(-1)
        preds = preds[:, :-1].reshape(-1)
        return metric.compute(predictions=preds, references=labels)

# %% ../nbs/03_model_lm_main.ipynb 12
def preprocess_lm_logits_for_metrics(logits, labels):
    if isinstance(logits, tuple):
        # Depending on the model and config, logits may contain extra tensors,
        # like past_key_values, but logits always come first
        logits = logits[0]
    return logits.argmax(dim=-1) 

# %% ../nbs/03_model_lm_main.ipynb 13
def finetune_lm(lr, # Learning rate
                bs, # Batch size
                wd, # Weight decay
                epochs, # Number of epochs
                ddict, # The HuggingFace datasetdict
                tokenizer,# HuggingFace tokenizer
                o_dir = './tmp_weights', # Directory to save weights
                save_checkpoint=False, # Whether to save weights (checkpoints) to o_dir
                model=None, # NLP model
                model_init=None, # A function to initialize model
                data_collator=None, # HuggingFace data collator
                compute_metrics=None, # A function to compute metric, default to `compute_lm_accuracy`
                grad_accum_steps=2, # The batch at each step will be divided by this integer and gradient will be accumulated over gradient_accumulation_steps steps.
                lr_scheduler_type='cosine',  # The scheduler type to use. Including: linear, cosine, cosine_with_restarts, polynomial, constant, constant_with_warmup
                warmup_ratio=0.1, # The warmup ratio for some lr scheduler
                no_valid=False, # Whether there is a validation set or not
                seed=None, # Random seed
                report_to='none', # The list of integrations to report the results and logs to. Supported platforms are "azure_ml", "comet_ml", "mlflow", "neptune", "tensorboard","clearml" and "wandb". Use "all" to report to all integrations installed, "none" for no integrations.
                trainer_class=None, # You can include the class name of your custom trainer here
            ):
    "The main model training/finetuning function"
    torch.cuda.empty_cache()
    gc.collect()
    
    if seed:
        seed_everything(seed)
        
    training_args = TrainingArguments(o_dir, 
                                     learning_rate=lr, 
                                     warmup_ratio=warmup_ratio,
                                     lr_scheduler_type=lr_scheduler_type, 
                                     fp16=True,
                                     do_train=True,
                                     do_eval= not no_valid,
                                     evaluation_strategy="no" if no_valid else "epoch", 
                                     save_strategy="epoch" if save_checkpoint else 'no',
                                     overwrite_output_dir=True,
                                     gradient_accumulation_steps=grad_accum_steps,
                                     per_device_train_batch_size=bs, 
                                     per_device_eval_batch_size=bs,
                                     num_train_epochs=epochs, weight_decay=wd,
                                     report_to=report_to,
                                     logging_dir=os.path.join(o_dir, 'log') if report_to!='none' else None,
                                     logging_steps = len(ddict["train"]) // bs,
                                     )

    # instantiate trainer
    trainer_class = Trainer if trainer_class is None else trainer_class
    trainer = trainer_class(
        model=model,
        model_init=model_init if model is None else None,
        args=training_args,
        train_dataset=ddict['train'],#.shard(200, 0)
        eval_dataset=ddict['validation'] if not no_valid else None,
        data_collator=data_collator,
        tokenizer=tokenizer,
        compute_metrics=compute_metrics if not no_valid else None,
        preprocess_logits_for_metrics=preprocess_lm_logits_for_metrics if not no_valid else None
    )
    
    trainer.train()
    return trainer

# %% ../nbs/03_model_lm_main.ipynb 15
def extract_hidden_states(batch,
                          model=None, # NLP model
                          model_input_names=['input_ids', 'token_type_ids', 'attention_mask'], # Model required inputs, from tokenizer.model_input_names
                          data_collator=None, # HuggingFace data collator
                          state_name='last_hidden_state', # Name of the state to extract
                          state_idx=0, # The index (or indices) of the state to extract
                          device = None, # device that the model is trained on
                          ):
    state_idx = val2iterable(state_idx)
    
    if data_collator is not None:    
# --- Convert from  
# {'input_ids': [tensor([    0, 10444,   244, 14585,   125,  2948,  5925,   368,     2]), 
#                tensor([    0, 16098,  2913,   244,   135,   198, 34629,  6356,     2])]
# 'attention_mask': [tensor([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]), 
#                    tensor([1, 1, 1, 1, 1, 1, 1, 1, 1])]
#                    }
# --- to
# [{'input_ids': tensor([    0, 10444,   244, 14585,   125,  2948,  5925,   368,     2]),
#   'attention_mask': tensor([1, 1, 1, 1, 1, 1, 1, 1, 1])},
#  {'input_ids': tensor([    0, 16098,  2913,   244,   135,   198, 34629,  6356,     2]),
#   'attention_mask': tensor([1, 1, 1, 1, 1, 1, 1, 1, 1])}]

        # remove string text, due to transformer new version       
        collator_inp = []
        ks = [k for k in batch.keys() if k in model_input_names]
        vs = [batch[k] for k in ks]
        for pair in zip(*vs):
            collator_inp.append({k:v for k,v in zip(ks,pair)})
        
        batch = data_collator(collator_inp)
    
    inputs = {k:v.to(device) for k,v in batch.items()
              if k in model_input_names}
            
    # switch to eval mode for evaluation
    if model.training:
        model.eval()
    with torch.no_grad():
        outputs = model(**inputs)
        states = outputs[state_name]
        for i in state_idx:
            if isinstance(states,tuple):
                states = states[i]
            else:
                states = states[:,i]
    # Switch back to train mode
    if not model.training:
        model.train()

    return {state_name:states.cpu().numpy()}

# %% ../nbs/03_model_lm_main.ipynb 16
class ModelLMController():
    def __init__(self,
                 model, # NLP language model
                 data_store=None, # a TextDataLMController/TextDataLMControllerStreaming object
                 seed=None, # Random seed
                ):
        self.model = model
        self.data_store = data_store
        self.seed = seed
        
    def fit(self,
            epochs, # Number of epochs
            learning_rate, # Learning rate
            ddict=None, # DatasetDict to fit (will override data_store)
            compute_metrics=None, # A function to compute metric, default to `compute_lm_accuracy`
            batch_size=16, # Batch size
            weight_decay=0.01, # Weight decay
            lr_scheduler_type='cosine', # The scheduler type to use. Including: linear, cosine, cosine_with_restarts, polynomial, constant, constant_with_warmup
            warmup_ratio=0.1, # The warmup ratio for some lr scheduler
            o_dir = './tmp_weights', # Directory to save weights
            save_checkpoint=False, # Whether to save weights (checkpoints) to o_dir
            hf_report_to='none', # The list of HuggingFace-allowed integrations to report the results and logs to
            grad_accum_steps=2, # Gradient will be accumulated over gradient_accumulation_steps steps.
            tokenizer=None, # Tokenizer (to override one in ```data_store```)
            data_collator=None, # Data Collator (to override one in ```data_store```)
            is_mlm=None, # Whether this is masked LM or casual LM
            trainer_class=None, # You can include the class name of your custom trainer here
           ):
        
        if tokenizer is None: tokenizer=check_and_get_attribute(self.data_store,'tokenizer')
        if data_collator is None: data_collator=check_and_get_attribute(self.data_store,'data_collator')
        if ddict is None: ddict = check_and_get_attribute(self.data_store,'main_ddict')
        if is_mlm is None: is_mlm = check_and_get_attribute(self.data_store,'is_mlm')
            
        if compute_metrics is None:
            compute_metrics=partial(compute_lm_accuracy,is_mlm=is_mlm)
        
        if len(set(ddict.keys()) & set(['train','training']))==0:
            raise ValueError("Missing the following key for DatasetDict: train/training")
        no_valid = len(set(ddict.keys()) & set(['validation','val','valid']))==0
        
        trainer = finetune_lm(learning_rate,
                              batch_size,
                              weight_decay,
                              epochs,
                              ddict,
                              tokenizer,
                              o_dir,
                              save_checkpoint=save_checkpoint,
                              model=self.model,
                              data_collator=data_collator,
                              compute_metrics=compute_metrics,
                              grad_accum_steps=grad_accum_steps,
                              lr_scheduler_type=lr_scheduler_type,
                              warmup_ratio=warmup_ratio,
                              no_valid=no_valid,
                              seed=self.seed,
                              trainer_class=trainer_class,
                              report_to=hf_report_to)
        
        if not no_valid:
            eval_results = trainer.evaluate()
            try:
                perplexity = math.exp(eval_results["eval_loss"])
                print(f'Perplexity on validation set: {perplexity:.3f}')
            except OverflowError:
                perplexity = float("inf")
                print(f'Perplexity on validation set: {perplexity}')
            eval_results['perplexity'] = perplexity
        
        self.trainer = trainer
        
    def predict_raw_text(self,
                         content:dict|list|str, # Either a single sentence, list of sentence or a dictionary where keys are metadata, values are list
                         print_result=True, # To whether print the result in readable format, or get the result returned
                         **kwargs, # keyword arguments for HuggingFace's text-generation (for clm) or fill-mask (for mlm)
                        ):
        # Example of kwargs for text-generation:
        # https://huggingface.co/docs/transformers/v4.33.2/en/main_classes/text_generation#transformers.GenerationMixin.generate
        
        if not isinstance(self.data_store,(TextDataLMController,TextDataLMControllerStreaming)) or not self.data_store._processed_call:
            raise ValueError('This functionality needs a TextDataController object which has processed some training data')
        test_dset = self.data_store.prepare_test_dataset_from_raws(content)
        result = self.predict_ddict(test_dset,**kwargs)
        if print_result:
            is_mlm = check_and_get_attribute(self.data_store,'is_mlm')
            for preds in result:
                for pred in preds:
                    if is_mlm:
                        print(f"Score: {pred['score']:.3f} >>> {pred['sequence']}")
                    else:
                        print(f">>> {pred['generated_text']}")
                print('-'*20)
        else:
            return result
    
    def predict_ddict(self,
                      dset:Dataset, # A processed and tokenized Dataset
                      **kwargs, # keyword arguments for HuggingFace's text-generation (for clm) or fill-mask (for mlm)
                     ):
        is_mlm = check_and_get_attribute(self.data_store,'is_mlm')
        tokenizer=check_and_get_attribute(self.data_store,'tokenizer')
        main_text=check_and_get_attribute(self.data_store,'main_text')
        _task = 'fill-mask' if is_mlm else 'text-generation'
        pipeline_obj = pipeline(_task,model=self.model,tokenizer=tokenizer,device=self.model.device)
        str_list = dset[main_text]
        if _task=='fill-mask':
            all_tfms = self.data_store.content_tfms 
            all_tfms = partial(func_all,functions=all_tfms) if len(all_tfms) else lambda x: x
            mask_str = all_tfms(tokenizer.mask_token)
            str_list = [str(s).replace(mask_str,tokenizer.mask_token) for s in str_list]
        return [pipeline_obj(s,**kwargs) for s in str_list]
    
    def get_hidden_states_from_raw_text(self,
                                        content:dict|list|str, # Either a single sentence, list of sentence or a dictionary where keys are metadata, values are list
                                        state_name, # Name of the (hidden) state to extract
                                        state_idx=0, # The index (or indices) of the state to extract. For `hidden_states`, accept multiple values
                                       ):
        if not isinstance(self.data_store,(TextDataLMController,TextDataLMControllerStreaming)) or not self.data_store._processed_call:
            raise ValueError('This functionality needs a TextDataController object which has processed some training data')
        dset = self.data_store.prepare_test_dataset_from_raws(content,do_tokenize=True)
        return self.get_hidden_states(dset,
                                      state_name=state_name,
                                      state_idx=state_idx,
                                      batch_size=1
                                     )

        
    def get_hidden_states_from_raw_dset(self,
                                        dset: Dataset, # A raw HuggingFace dataset
                                        state_name, # Name of the (hidden) state to extract
                                        state_idx=0, # The index (or indices) of the state to extract. For `hidden_states`, accept multiple values
                                        batch_size=16, # GPU batch size
                                       ):
        if not isinstance(self.data_store,(TextDataLMController,TextDataLMControllerStreaming)) or not self.data_store._processed_call:
            raise ValueError('This functionality needs a TextDataController object which has processed some training data')
        dset = self.data_store.prepare_test_dataset(dset,do_tokenize=True)
        return self.get_hidden_states(dset,
                                      state_name=state_name,
                                      state_idx=state_idx,
                                      batch_size=batch_size
                                     )
        
        
    def get_hidden_states(self,
                          ddict:DatasetDict|Dataset=None, # A processed and tokenized DatasetDict/Dataset (will override one in ```data_store```)
                          ds_type='test', # The split of DatasetDict to predict
                          state_name='last_hidden_state', # Name of the (hidden) state to extract
                          state_idx=0, # The index (or indices) of the state to extract. For `hidden_states`, accept multiple values
                          batch_size=16, # GPU batch size
                         ):
        
        tokenizer=check_and_get_attribute(self.data_store,'tokenizer')
        if ddict is None: ddict = check_and_get_attribute(self.data_store,'main_ddict')
        if isinstance(ddict,DatasetDict):
            if ds_type not in ddict.keys():
                raise ValueError(f'{ds_type} is not in the given DatasetDict')
            ddict = ddict[ds_type]
        
        ddict.set_format("torch",
                        columns=tokenizer.model_input_names)
        
        results = ddict.map(partial(extract_hidden_states,
                                 model=self.model,
                                 model_input_names = tokenizer.model_input_names,
                                 state_name=state_name,
                                 state_idx=state_idx,
                                 device = self.model.device),
                           batched=True,
                           batch_size=batch_size
                          )
        results.set_format('numpy')
        return results

